# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import Task, DBSession
__author__ = 'vahid'


class AbortCommand(Command):
    name = 'abort'
    description = 'Aborts currently active task'

    def do_job(self):
        active_task = Task.get_active_task()
        if active_task:
            print 'Aborting active task: %s' % active_task
            DBSession.delete(active_task)
            DBSession.commit()
        else:
            print "You don't have any active task"



# -*- coding: utf-8 -*-
import timesheet
from timesheet.commands import Command
__author__ = 'vahid'


class VersionCommand(Command):
    name = 'version'
    description = 'Prints the version'

    def do_job(self):
        print timesheet.__version__



# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import Task, DBSession
__author__ = 'vahid'


class EndCommand(Command):
    name = 'end'
    description = 'Ends a task'

    def do_job(self):
        active_task = Task.get_active_task()
        if not active_task:
            print "You don't have any active task"
            return

        active_task.end()
        DBSession.commit()
        print 'Task ended: %s' % active_task



# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import Task, DBSession
import argparse
__author__ = 'vahid'


class EditLastCommand(Command):
    name = 'edit-last'
    description = 'Edit last task'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('task', nargs=argparse.REMAINDER, default=[], help="The task name")

    def do_job(self):
        task = Task.get_last_task()
        if task:
            task.title = ' '.join(self.args.task)
            DBSession.commit()
        else:
            print 'Last task not found'



# -*- coding: utf-8 -*-
import sys
from timesheet.configuration import init_config, config
__author__ = 'vahid'
__version__ = '0.8.4'


def entrypoint():

    # initializing configuration
    init_config()

    # initializing models if not initialized yet
    from timesheet import models
    models.init()

    # Preparing cli arguments
    from timesheet import cli
    args = cli.parse_ars()

    # Dispatch and execute the command
    args.command_class(args).do_job()

    sys.exit(0)



# -*- coding: utf-8 -*-

from .base import Command, get_available_commands, get_available_command_names, get_command
from .help import HelpCommand
from .start import StartCommand
from .end import EndCommand
from .active import ActiveCommand
from .full_report import FullReportCommand
from .subjects import SubjectsCommand
from .import_ import ImportCommand
from .export import ExportCommand
from .edit_last import EditLastCommand
from .abort import AbortCommand
from .rename import RenameCommand
from .daily_report import DailyReportCommand
from .daily_detail import DailyDetailCommand
from .delete import DeleteCommand
from .version import VersionCommand

__all__ = ['Command']




class Command(object):
    """
    Abstract command class
    """
    name = ''
    description = ''
    parser = None

    def __init__(self, args):
        self.args = args

    @classmethod
    def create_parser(cls, subparsers):
        cls.parser = subparsers.add_parser(cls.name, help=cls.description)
        cls.parser.set_defaults(command_class=cls)
        cls.add_arguments()
        return cls.parser

    @classmethod
    def add_arguments(cls):
        pass

    @classmethod
    def help(cls):
        cls.parser.print_help()

    def do_job(self):
        raise NotImplementedError()


def __get_available_commands(command_class):
    commands = command_class.__subclasses__()
    for c in list(commands):
        commands.extend(__get_available_commands(c))
    return commands


def get_available_commands():
    return __get_available_commands(Command)


def get_available_command_names():
    return [c.name for c in get_available_commands()]


def get_command(command_name):
    for c in get_available_commands():
        if c.name == command_name:
            return c
    raise ValueError('Invalid command name: %s' % command_name)



# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import DBSession, Task, Subject
from timesheet.commands.completers import subject_completer
from prettytable import PrettyTable
from datetime import timedelta, datetime
from sqlalchemy import func

__author__ = 'gmc'


class DailyDetailCommand(Command):
    name = 'daily-detail'
    description = 'Print detailed daily report'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('days', nargs='?', default=30, type=int, help="Days, integer, default: 30")

    def do_job(self):
        """
        SELECT *, coalesce(end_time,now)-start_time
        FROM task t, subject s
        WHERE t.subject_id=s.id AND start_time > GetDate() - @days
        ORDER BY start_time
        :return:
        """
        session = DBSession()
        now = datetime.now()
        time_worked = (func.julianday(func.coalesce(Task.end_time,now)) - func.julianday(Task.start_time)) * 86400

        query = session.query(Task.start_time,
                              func.coalesce(Task.end_time, now),
                              time_worked,
                              Subject.title,
                              Task.title) \
            .filter(Subject.id==Task.subject_id) \
            .filter(func.date(Task.start_time) > func.date('now', '-%s day' % self.args.days)) \
            .order_by(Task.start_time)

        print '\n'

        table = PrettyTable(['Start', 'End', 'Time', 'Subject', 'Title'])
        table.align["Title"] = "l"

        total_time = 0
        day_total = 0
        last_date = None

        for row in query:
            if last_date == None:
                last_date = row[0].date()

            if row[0].date() != last_date:
                table.add_row([
                    '', '', timedelta(seconds=round(day_total)), '', ''
                ])
                last_date = row[0].date()
                day_total = 0

            day_total += row[2]
            total_time += row[2]

            table.add_row([
                row[0],
                row[1],
                timedelta(seconds=round(row[2])),
                row[3],
                row[4],
            ])

        if day_total > 0:
            table.add_row([
                '', '', timedelta(seconds=round(day_total)), '', ''
            ])

        print table
        print 'Total Work time: %s' % timedelta(seconds=total_time)
        print



# -*- coding: utf-8 -*-
import sys
import csv
from timesheet.commands import Command
from timesheet import config
from timesheet.models import Subject, DBSession, Task
from datetime import datetime
__author__ = 'vahid'


class ImportCommand(Command):
    name = 'import'
    description = 'Import csv file with columns: subject, task, start, end'

    def __init__(self, *args, **kw):
        super(ImportCommand, self).__init__(*args, **kw)
        self.subjects = {}

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('input', nargs='?', help="Input filename. if omitted, the standard input will be used")

    def process_row(self, row):
        subject_name, task_name, start_time, end_time = row
        start_time = datetime.strptime(start_time, config.datetime_format)
        end_time = datetime.strptime(end_time, config.datetime_format)

        if not subject_name in self.subjects:
            self.subjects[subject_name] = Subject.ensure(subject_name)
            DBSession.commit()

        task = Task(title=task_name, start_time=start_time, end_time=end_time)
        self.subjects[subject_name].tasks.append(task)
        print 'Adding %s' % task
        DBSession.commit()

    def do_job(self):

        if self.args.input:
            csv_file = open(self.args.input)
        else:
            csv_file = sys.stdin

        try:
            reader = csv.reader(csv_file)
            for row in reader:
                self.process_row(row)
        finally:
            if csv_file and csv_file != sys.stdin:
                csv_file.close()



# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import Subject
__author__ = 'vahid'


class SubjectsCommand(Command):
    name = 'subjects'
    description = 'Print all subjects'

    def do_job(self):
        print '\n'.join(Subject.all_titles())



# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import DBSession, Task
from timesheet.commands.completers import subject_completer
from prettytable import PrettyTable
from datetime import timedelta, datetime
from sqlalchemy import func

__author__ = 'vahid'


class DailyReportCommand(Command):
    name = 'daily-report'
    description = 'Print daily report'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('days', nargs='?', default=30, type=int, help="Days, integer, default: 30")

    def do_job(self):
        """
        SELECT sum(t.end_time - t.start_time)
        FOM task t
        GROUP BY day(t.start_time)
        WHERE start_time > GetDate() - @days
        :return:
        """
        session = DBSession()
        daywork = func.sum(func.julianday(Task.end_time) - func.julianday(Task.start_time)) * 86400
        day = func.date(Task.start_time)
        query = session.query(day, daywork) \
            .group_by(day) \
            .filter(func.date(Task.start_time) > func.date('now', '-%s day' % self.args.days)) \
            .filter(Task.end_time != None) \
            .order_by(Task.start_time)

        print '\n'

        table = PrettyTable(['Day', 'Work Time', 'Graph'])
        table.align["Graph"] = "l"
        total_hours = timedelta(0)
        last_day = None
        for row in query:
            day = datetime.strptime(row[0], '%Y-%m-%d').date()
            if last_day:
                diff = (day - last_day)
                for i in range(diff.days - 1):
                    table.add_row([last_day + timedelta(i + 1), 0, ''])

            worktime = timedelta(seconds=round(row[1]))
            total_hours += worktime
            table.add_row([day, worktime, '#' * int(round((row[1] * 60 / 86400)))])
            last_day = day

        print table
        print 'Total Work time: %s' % total_hours
        print



# -*- coding: utf-8 -*-
from timesheet.models import Subject
__author__ = 'vahid'


# noinspection PyUnusedLocal
def subject_completer(prefix, **kwargs):
    try:
        return [c for c in Subject.all_titles() if c.startswith(prefix)]
    except Exception as ex:
        return [str(ex)]



# -*- coding: utf-8 -*-
from timesheet.commands.completers import subject_completer
from timesheet.commands import Command
from timesheet.models import Subject, DBSession
__author__ = 'vahid'


class DeleteCommand(Command):
    name = 'delete'
    description = 'Deletes a subject '

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('subject', nargs='?', help="Subject to delete.")\
            .completer = subject_completer

    def do_job(self):
        subject = Subject.query.filter(Subject.title == self.args.subject).first()
        if not subject:
            print 'Subject can not found: %s' % self.args.subject
        else:
            DBSession.delete(subject)
            DBSession.commit()

        print "Subject `%s` was deleted successfully" % self.args.subject



# -*- coding: utf-8 -*-
from pymlconf import ConfigManager
from os import path, mkdir
from appdirs import user_data_dir, user_config_dir

__author__ = 'vahid'

data_dir = path.abspath(path.join(user_data_dir(), 'timesheet'))
user_config_file = path.abspath(path.join(user_config_dir(), 'timesheetrc'))

__builtin_config__ = """
db:
  uri: sqlite:///%(data_dir)s/timesheet.sqlite
  echo: false

datetime_format: %(time_format)s
date_format: %(date_format)s

"""

_cfg = None


def init_config(config_file=None):
    global _cfg
    if not path.exists(data_dir):
        mkdir(data_dir)

    config_files = [user_config_file]
    if config_file:
        config_files.append(config_file)

    _cfg = ConfigManager(
        __builtin_config__,
        files=config_files,
        context=dict(
            data_dir=data_dir,
            time_format='"%Y-%m-%d %H:%M"',
            date_format='"%Y-%m-%d"')
    )


class ConfigProxy(object):

    def __getattr__(self, key):
        if _cfg is None:
            raise Exception("Configuration not initialized yet.")
        return getattr(_cfg, key)

    def __setattr__(self, key, value):
        if _cfg is None:
            raise Exception("Configuration not initialized yet.")
        setattr(_cfg, key, value)


config = ConfigProxy()


# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import Subject, DBSession
from timesheet.commands.completers import subject_completer
__author__ = 'vahid'


class RenameCommand(Command):
    name = 'rename'
    description = 'Renames a subject'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('subject', help="Subject to do something about that.").completer = subject_completer
        cls.parser.add_argument('new_name', help="Subject's new name").completer = subject_completer

    def do_job(self):
        if not self.args.subject:
            print "Please specify a subject to rename"

        subject = Subject.query.filter(Subject.title == self.args.subject).first()
        if not subject:
            print 'Subject can not found: %s' % self.args.subject
            return

        new_subject = Subject.query.filter(Subject.title == self.args.new_name).first()
        if new_subject:
            print "There is already a subject with name: %s" % self.args.new_name
            answer = raw_input("Do you want to merge these subjects? [Y/n]: ")
            if not answer or answer.lower() == 'y':
                for task in subject.tasks:
                    task.subject = new_subject
                DBSession.flush()
                DBSession.delete(subject)
                DBSession.commit()
            else:
                print 'Operation aborted by user.'

        else:
            subject.title = self.args.new_name
            DBSession.commit()





# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import Subject, Task
from timesheet.commands.completers import subject_completer
from prettytable import PrettyTable
from datetime import timedelta, date
__author__ = 'vahid'


class FullReportCommand(Command):
    name = 'report'
    description = 'Print report about an subject or all subjects'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('subject', nargs='?', help="Subject to do something about that.")\
            .completer = subject_completer
        cls.parser.add_argument('--today', action='store_true', help="Only prints today's tasks.")

    def report_subject(self, subject):

        print '\n'
        print '%s' % subject.title

        table = PrettyTable(['Task Name', 'Duration', 'From', 'To'])
        table.align["Task Name"] = "l"
        table.align["Duration"] = "c"
        table.padding_width = 1  # default

        q = Task.query.filter(Task.subject == subject)
        if self.args.today:
            q = q.filter(Task.start_time > date.today())

        total_duration = timedelta()
        for task in q:
            total_duration += task.duration
            table.add_row(['' if not task.title else task.title,
                           task.duration_formatted,
                           task.start_time_string,
                           task.end_time_string])

        print table
        ts = total_duration.total_seconds()
        h = ts / 3600
        m = (ts % 3600) / 60
        print 'Total: %.2d:%.2d' % (h, m)
        print

    def do_job(self):
        if self.args.subject:
            subject = Subject.query.filter(Subject.title == self.args.subject).first()
            if not subject:
                print 'Subject can not found: %s' % self.args.subject
            else:
                self.report_subject(subject)
        else:
            for subject in Subject.query.order_by(Subject.entry_time):
                self.report_subject(subject)



# -*- coding: utf-8 -*-
from timesheet.commands import Command
from timesheet.models import Subject, Task, DBSession
from timesheet.commands.completers import subject_completer
import argparse
__author__ = 'vahid'


class StartCommand(Command):
    name = 'start'
    description = 'Starts a task'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('subject', help="Subject to do something about that.").completer = subject_completer
        cls.parser.add_argument('task', nargs=argparse.REMAINDER, metavar='task', default=[], choices=[],
                                help="Task title.")

    def do_job(self):
        active_task = Task.get_active_task()
        if active_task:
            print 'You have an active task: %s' % active_task
            answer = raw_input("Do you want to terminate the currently active task ([y]/n)? ")
            if not answer or answer.lower() == 'y':
                active_task.end()
            else:
                return

        subject = Subject.ensure(self.args.subject)
        task = Task(title=' '.join(self.args.task))
        subject.tasks.append(task)

        DBSession.commit()
        print 'Started task: %s' % task





# -*- coding: utf-8 -*-
import sys
import csv
from timesheet.commands import Command
from timesheet.models import Task, Subject
from timesheet.commands.completers import subject_completer
__author__ = 'vahid'


class ExportCommand(Command):
    name = 'export'
    description = 'Export the database into csv file with columns: subject, task, start, end'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('output', nargs='?', help="Output filename. if omitted, \
        the standard output will be used")
        cls.parser.add_argument('-s', '--subject',
                                action='append',
                                default=[],
                                help="Subject to filter the result ")\
            .completer = subject_completer

    def do_job(self):

        if self.args.output:
            csv_file = open(self.args.output, 'w')
        else:
            csv_file = sys.stdout

        try:
            query = Task.query.join(Task.subject).order_by(Subject.entry_time, Task.start_time)
            if self.args.subject:
                query = query.filter(Subject.title.in_(self.args.subject))

            writer = csv.writer(csv_file)
            for task in query:
                writer.writerow([task.subject.title, task.title, task.start_time_string, task.end_time_string])
        finally:
            if csv_file and csv_file != sys.stdout:
                csv_file.close()



# -*- coding: utf-8 -*-
from timesheet.commands import Command, get_command, get_available_command_names
from argcomplete.completers import ChoicesCompleter
__author__ = 'vahid'


class HelpCommand(Command):
    name = 'help'
    description = 'Prints help for given command'

    @classmethod
    def add_arguments(cls):
        cls.parser.add_argument('command', nargs='?', help="Command to print help about that").completer = \
            ChoicesCompleter(get_available_command_names())

    def do_job(self):
        if self.args.command:
            command_class = get_command(self.args.command)
            command_class.help()
        else:
            from timesheet.cli import parser
            parser.print_help()


# -*- coding: utf-8 -*-
import argparse
import argcomplete
from timesheet.commands import get_available_commands
__author__ = 'vahid'


parser = argparse.ArgumentParser(description='Simple timesheet system.')
subparsers = parser.add_subparsers(title='Commands', description='')

for cmd in get_available_commands():
    cmd.create_parser(subparsers)

argcomplete.autocomplete(parser)


def parse_ars():
    # args, _reminder = parser.parse_known_args()
    args = parser.parse_args()
    return args



# -*- coding: utf-8 -*-
import os
import re
from setuptools import find_packages
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# reading cherryblog.py version without importing any modules from that package
with open(os.path.join(os.path.dirname(__file__), 'timesheet', '__init__.py')) as v_file:
    package_version = re.compile(r".*__version__ = '(.*?)'", re.S).match(v_file.read()).group(1)

dependencies = [
    'sqlalchemy',
    'pymlconf>=0.3.10',
    'appdirs>=1.3.0',
    'argcomplete',
    'PrettyTable'
]


def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()


setup(
    name="timesheet",
    version=package_version,
    author="Vahid Mardani",
    author_email="vahid.mardani@gmail.com",
    url="http://github.com/pylover/timesheet",
    description="Simple timesheet tracking system",
    packages=find_packages(exclude=['ez_setup']),
    platforms=["any"],
    long_description=read('README.rst'),
    install_requires=dependencies,
    dependency_links=[
        'git+https://github.com/pylover/elixir.git#egg=elixir-0.8.1'
    ],
    entry_points={
        'console_scripts': [
            'timesheet = timesheet:entrypoint'
        ]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: Freeware",
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

)



# -*- coding: utf-8 -*-
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker, relationship, backref
from sqlalchemy import create_engine, Column, DateTime, String, Integer, ForeignKey
from datetime import datetime, timedelta
from timesheet import config
__author__ = 'vahid'

maker = sessionmaker()
DBSession = scoped_session(maker)
BaseModel = declarative_base()
BaseModel.query = DBSession.query_property()
metadata = BaseModel.metadata


class Subject(BaseModel):
    __tablename__ = 'subject'

    id = Column(Integer, primary_key=True)
    title = Column(String(50), nullable=False, index=True, unique=True)
    entry_time = Column(DateTime, nullable=False, default=datetime.now)

    @classmethod
    def ensure(cls, title):
        s = cls.query.filter(cls.title == title).first()
        if not s:
            s = cls(title=title)
            DBSession.add(s)
        return s

    @classmethod
    def all_titles(cls):
        return [s.title for s in cls.query.order_by(cls.title)]

    def __repr__(self):
        return '<%s created=%s >' % (
            self.title,
            self.entry_time.strftime(config.datetime_format)
        )


class Task(BaseModel):
    __tablename__ = 'task'

    id = Column(Integer, primary_key=True)
    title = Column(String(100), nullable=True)
    start_time = Column(DateTime, nullable=False, default=datetime.now)
    end_time = Column(DateTime, nullable=True)
    subject_id = Column(Integer, ForeignKey('subject.id'))

    subject = relationship("Subject", backref=backref('tasks', order_by=start_time))

    @classmethod
    def get_active_task(cls):
        task = cls.query.filter(cls.end_time == None).order_by(cls.start_time.desc()).first()
        return task

    @classmethod
    def get_last_task(cls):
        task = cls.query.order_by(cls.id.desc()).first()
        return task

    def end(self):
        self.end_time = datetime.now()

    @property
    def duration(self):
        if not self.end_time:
            return timedelta()
        else:
            return self.end_time - self.start_time

    @property
    def duration_formatted(self):
        total_seconds = self.duration.total_seconds()
        hours, remainder = divmod(total_seconds, 3600)
        return '%02d:%02d' % (hours, remainder / 60)

    def __repr__(self):
        return '\n%-12s%s\n%-12s%s\n%-12s%s\n%-12s%s\n%-12s%s' % (
            'Subject', "" if not self.subject else self.subject.title,
            'Title', self.title,
            'Start', self.start_time_string,
            'End', self.end_time_string,
            'Duration', self.duration_formatted
        )

    @property
    def start_time_string(self):
        return self.start_time.strftime(config.datetime_format)

    @property
    def end_time_string(self):
        return '' if not self.end_time else self.end_time.strftime(config.datetime_format)


def init():

    if DBSession.bind:
        return
    elif DBSession.registry.has():
        DBSession.registry.clear()

    engine = create_engine(config.db.uri, echo=config.db.echo)
    DBSession.configure(bind=engine)
    BaseModel.metadata.create_all(engine)
